#ifndef	__SITECING_MAGIC_H
#define __SITECING_MAGIC_H

/**
 * @file
 * @brief The magic numbers globally defined.
 */

namespace sitecing {

    /**
     * The magic numbers enumeration.
     */
    enum {
	/**
	 * There is no magic.
	 */
	__magic_none = 0,
	/**
	 * Here is where user-defined magic starts.
	 */
	__user_magical_numbers_start = 1,
	/**
	 * Here is where site-C-ing defined magic starts.
	 */
	__sitecing_magical_numbers_start = 0x8000,
	/**
	 * The compiler error occured. The parameters passed are:
	 *
	 * char *message, char *root_source, char *root_intermediate, char *root_so, char *component
	 */
	__magic_compile_error,
	/**
	 * The preprocessor error occured. The parameters passed are:
	 *
	 * char *message, char *root_source, char *root_intermediate, char *root_so, char *component,
	 * int line_number
	 */
	__magic_preprocess_error,
	/**
	 * Exception caught while executing the component. The parameters passed are:
	 *
	 * char *message, char *root_source, char *root_intermediate, char *root_so, char *component,
	 * const exception *exception_caught
	 */
	__magic_generic_exception,
	/**
	 * The component called as an action handler. The parameters passed are:
	 *
	 * char *root_source, char *root_intermediate, char *root_so, list<string>* args
	 */
	__magic_action,
	/**
	 * The component called as an HTTP status handler. The parameters passed are:
	 *
	 * char *root_source, char *root_intermediate, char *root_so, char *component,
	 * const http_status *http_status_caught
	 */
	__magic_http_status
    };
}

#endif /* __SITECING_MAGIC_H */
